/*-
 * %%----------------------------------------------------------------------------------------------
 * DLCM Technology - DLCM Admin - UserSynchronizationService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.service;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.auth.client.service.AuthorizationClientProperties;
import ch.unige.solidify.auth.model.AuthUserDto;
import ch.unige.solidify.auth.model.PersonInfo;
import ch.unige.solidify.auth.model.UserInfo;
import ch.unige.solidify.auth.service.UserInfoLocalService;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.TrustedRestClientTool;

public abstract class AbstractUserSynchronizationService {
  private static final String USERS_PATH = "/users/";
  private static final String CHANGE_ORCID_ENDPOINT = "/change-orcid";
  private static final Logger log = LoggerFactory.getLogger(AbstractUserSynchronizationService.class);
  private final UserInfoLocalService userInfoLocalService;
  private final TrustedRestClientTool trustedRestClientTool;
  private final TrustedRestClientService trustedRestClientService;
  private final String authModuleUrl;

  protected AbstractUserSynchronizationService(UserInfoLocalService userInfoLocalService, TrustedRestClientTool trustedRestClientTool,
          TrustedRestClientService trustedRestClientService, AuthorizationClientProperties authProperties) {
    this.userInfoLocalService = userInfoLocalService;
    this.trustedRestClientTool = trustedRestClientTool;
    this.trustedRestClientService = trustedRestClientService;
    this.authModuleUrl = authProperties.getAuthorizationServerUrl();
  }

  @Scheduled(cron = "${solidify.userSynchroCron: 0 59 23 * * *}")
  @Transactional
  public void synchronizeUsers() {
    try {
      // Update login info & role
      this.synchronizeUsersInfoFromAuthServer();
      // Synchronize Orcid
      this.synchronizeUsersOrcidWithAuthServer();
      log.info("Users synchronized successfully");
    } catch (RuntimeException e) {
      log.error("Users not synchronized");
    }
  }

  public void synchronizeUsersInfoFromAuthServer() {
    for (UserInfo userInfo : this.userInfoLocalService.findAll()) {
      try {
        this.updateUser(userInfo);
      } catch (RuntimeException e) {
        log.error("user information can't be updated from authorization server for user {}", userInfo.getResId());
      }
    }
  }

  public void synchronizeUsersOrcidWithAuthServer() {
    for (UserInfo userInfo : this.userInfoLocalService.findAll()) {
      try {
        this.updateUserOrcidOnAuthorizationServer(userInfo);
      } catch (RuntimeException e) {
        log.error("user information can't be updated to authorization server for user {}", userInfo.getResId());
      }
    }
  }
  public void updateUserOrcidOnAuthorizationServer(UserInfo userInfo) {
    final AuthUserDto authUserDto = this.getAuthenticatedUser(userInfo.getExternalUid());
    final PersonInfo personInfo = userInfo.getPerson();
    if (StringTool.isNullOrEmpty(authUserDto.getOrcid()) && !StringTool.isNullOrEmpty(personInfo.getOrcid())) {
      // Update user ORCID in Authorization
      AuthUserDto orcidInfo = new AuthUserDto();
      orcidInfo.setOrcid(personInfo.getOrcid());
      orcidInfo.setVerifiedOrcid(personInfo.isVerifiedOrcid());
      this.updateUserOrcidOnAuthServer(authUserDto.getExternalUid(), orcidInfo);
      log.info("User {}: ORCID updated with {}", authUserDto.getExternalUid(), personInfo.getOrcid());
    }
  }

  public AuthUserDto getAuthenticatedUser(String externalUid) {
    return this.trustedRestClientService.getResource(this.authModuleUrl + USERS_PATH + externalUid, AuthUserDto.class);
  }
  public void updateUser(UserInfo userInfo) {
    final AuthUserDto authUserDto = this.getAuthenticatedUser(userInfo.getExternalUid());
    this.userInfoLocalService.saveAuthUserDto(userInfo, authUserDto);
  }

  private void updateUserOrcidOnAuthServer(String externalUid, AuthUserDto orcidInfo) {
    RestTemplate restTemplate = this.trustedRestClientTool.getClient();
    AuthUserDto authUserDto = restTemplate.postForObject(this.authModuleUrl + USERS_PATH + externalUid + CHANGE_ORCID_ENDPOINT, orcidInfo,
            AuthUserDto.class);
    if (authUserDto == null) {
      throw new IllegalStateException();
    }
  }
}
