/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Index - ElasticsearchSettingsService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.index.settings.elasticsearch;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.elasticsearch.ElasticsearchStatusException;
import org.elasticsearch.action.admin.indices.delete.DeleteIndexRequest;
import org.elasticsearch.action.admin.indices.settings.get.GetSettingsRequest;
import org.elasticsearch.action.admin.indices.settings.get.GetSettingsResponse;
import org.elasticsearch.action.support.master.AcknowledgedResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.client.indices.CreateIndexRequest;
import org.elasticsearch.client.indices.CreateIndexResponse;
import org.elasticsearch.client.indices.GetMappingsRequest;
import org.elasticsearch.client.indices.GetMappingsResponse;
import org.elasticsearch.client.indices.PutMappingRequest;
import org.elasticsearch.cluster.metadata.MappingMetadata;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.rest.RestStatus;
import org.elasticsearch.xcontent.XContentType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Profile;
import org.springframework.core.io.ClassPathResource;
import org.springframework.stereotype.Service;

import com.carrotsearch.hppc.cursors.ObjectObjectCursor;

import ch.unige.solidify.IndexDefinition;
import ch.unige.solidify.IndexDefinitionList;
import ch.unige.solidify.IndexProperties;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.controller.IndexController;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.index.ElasticsearchClientProvider;
import ch.unige.solidify.index.settings.IndexingSettingsService;
import ch.unige.solidify.model.index.IndexSettings;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.JSONTool;
import ch.unige.solidify.util.StringTool;

@Profile("index-elasticsearch")
@Service
@ConditionalOnBean(IndexController.class)
public class ElasticsearchSettingsService extends IndexingSettingsService<IndexSettings> {
  private static final Logger log = LoggerFactory.getLogger(ElasticsearchSettingsService.class);

  private static final String MAPPING_HOME = "mapping";

  private final RestHighLevelClient esClt;
  private final int fieldsLimit;
  private String indexPrefix;
  private final int replicaNumber;
  private final int shardNumber;

  public ElasticsearchSettingsService(
          IndexProperties indexProperties,
          ElasticsearchClientProvider elasticsearchClientProvider) {
    this.indexPrefix = indexProperties.getIndexPrefix();
    this.esClt = elasticsearchClientProvider.getElasticsearchClient();
    this.shardNumber = indexProperties.getConfig().getShardNumber();
    this.replicaNumber = indexProperties.getConfig().getReplicaNumber();
    this.fieldsLimit = indexProperties.getConfig().getFieldsLimit();
  }

  @Override
  public boolean delete(IndexSettings ai) {
    if (!this.checkIndexName(ai.getResId())) {
      return false;
    }
    final DeleteIndexRequest request = new DeleteIndexRequest(ai.getResId());
    final AcknowledgedResponse res;
    try {
      res = this.esClt.indices().delete(request, RequestOptions.DEFAULT);
      return (res.isAcknowledged());
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error while deleting index " + ai.getResId(), e);
    }
  }

  @Override
  public List<IndexSettings> findAll(IndexSettings search) {
    final List<IndexSettings> list = new ArrayList<>();
    final GetSettingsRequest request = new GetSettingsRequest();
    final GetSettingsResponse res;
    try {
      res = this.esClt.indices().getSettings(request, RequestOptions.DEFAULT);
      for (final ObjectObjectCursor<String, Settings> cursor : res.getIndexToSettings()) {
        if (this.checkIndexName(cursor.key)) {
          final IndexSettings ai = new IndexSettings();
          ai.setResId(cursor.key);
          ai.setSettings(this.getSettings(cursor.value));
          list.add(ai);
        }
      }
      return list;
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error while running search " + search + " on index " + this.indexPrefix, e);
    }
  }

  @Override
  public IndexSettings findOne(String index) {
    if (!this.checkIndexName(index)) {
      log.error("Wrong index name ({})", index);
      return null;
    }
    final IndexSettings ai = new IndexSettings();
    ai.setResId(index);
    // Get index settings
    GetSettingsResponse resS;
    try {
      final GetSettingsRequest request = new GetSettingsRequest().indices(index);
      resS = this.esClt.indices().getSettings(request, RequestOptions.DEFAULT);
    } catch (final ElasticsearchStatusException e) {
      if (RestStatus.NOT_FOUND.equals(e.status())) {
        log.warn("Index '{}' does not exist", index);
        return null;
      }
      throw e;
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error while searching on index " + index, e);
    }
    if (resS.getIndexToSettings().values().isEmpty()) {
      return null;
    }
    final Map<String, Object> indexDetails = this.getSettings(resS.getIndexToSettings().get(index));
    // Get index mapping
    final GetMappingsRequest request = new GetMappingsRequest().indices(index);
    final GetMappingsResponse resM;
    try {
      resM = this.esClt.indices().getMapping(request, RequestOptions.DEFAULT);
      for (final String mappingIndexName : resM.mappings().keySet()) {
        final MappingMetadata typeMapping = resM.mappings().get(mappingIndexName);
        indexDetails.putAll(typeMapping.getSourceAsMap());
      }
      ai.setSettings(indexDetails);
      return ai;
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error while searching on index " + index, e);
    }

  }

  @Override
  public void init(IndexDefinitionList indexList) {
    for (IndexDefinition index : indexList.getIndexList()) {
      this.createIndex(index);
    }
  }

  @Override
  public IndexSettings save(IndexSettings index) {

    if (!this.checkIndexName(index.getResId())) {
      throw new SolidifyRuntimeException("Wrong index name (" + index.getResId() + ")");
    }

    Settings.Builder builder = Settings.builder();
    if (!index.getSettings().isEmpty()) {
      builder.loadFromMap(index.getSettings());
    }
    builder.put("index.number_of_shards", this.shardNumber);
    builder.put("index.number_of_replicas", this.replicaNumber);
    builder.put("index.mapping.total_fields.limit", this.fieldsLimit);

    final CreateIndexRequest request = new CreateIndexRequest(index.getResId());
    request.settings(builder);

    try {
      final CreateIndexResponse res = this.esClt.indices().create(request, RequestOptions.DEFAULT);
      if (res.isAcknowledged()) {
        return index;
      } else {
        throw new SolidifyRuntimeException("Cannot create index (" + index + ")");
      }
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Error saving index " + index.getResId(), e);
    }
  }

  @Override
  public IndexSettings update(IndexSettings t) {
    throw new UnsupportedOperationException("Index Setting update");
  }

  private boolean checkIndexName(String indexName) {
    return indexName.startsWith(this.indexPrefix);
  }

  private void createIndex(IndexDefinition indexDefinition) {
    if (this.findOne(indexDefinition.getIndexName()) == null) {
      final IndexSettings index = new IndexSettings();
      index.setResId(indexDefinition.getIndexName());
      try {
        this.setSettings(index, indexDefinition);
        this.save(index);
        log.info("Index '{}' created", indexDefinition.getIndexName());
        this.setMapping(indexDefinition.getIndexName(), indexDefinition.getIndexMapping());
        log.info("Index '{}' mapping defined", indexDefinition.getIndexName());
      } catch (final SolidifyRuntimeException e) {
        final String message = "Cannot create '" + indexDefinition.getIndexName() + "' index";
        log.error(message);
        throw new SolidifyRuntimeException(message, e);
      } catch (final IOException e) {
        final String message = "Cannot create mapping for '" + indexDefinition.getIndexName() + "' index";
        log.error(message);
        throw new SolidifyRuntimeException(message, e);
      }
    }
  }

  private String getMapping(String indexMapping) throws IOException {
    if (!StringTool.isNullOrEmpty(indexMapping)) {
      return this.getMappingData(indexMapping);
    }
    return null;
  }

  private String getMappingData(String mapping) throws IOException {
    final ClassPathResource mappingResource = new ClassPathResource(MAPPING_HOME + SolidifyConstants.URL_SEP + mapping);
    return FileTool.toString(mappingResource.getInputStream());
  }

  private Map<String, Object> getSettings(Settings settings) {
    final Map<String, Object> list = new HashMap<>();
    for (final String setting : settings.names()) {
      if (settings.hasValue(setting)) {
        list.put(setting, settings.get(setting));
      } else {
        list.put(setting, this.getSettings(settings.getAsSettings(setting)));
      }
    }
    return list;
  }

  private boolean setMapping(String indexName, String indexMapping) throws IOException {
    final String mapping = this.getMapping(indexMapping);
    if (mapping != null) {
      final PutMappingRequest putMappingRequest = new PutMappingRequest(indexName)
              .source(mapping, XContentType.JSON);
      final AcknowledgedResponse res = this.esClt.indices().putMapping(putMappingRequest, RequestOptions.DEFAULT);
      return res.isAcknowledged();
    }
    return true;
  }

  private void setSettings(IndexSettings indexSettings, IndexDefinition indexDefinition) throws IOException {
    if (!StringTool.isNullOrEmpty(indexDefinition.getIndexSettings())) {
      final ClassPathResource settingsResource = new ClassPathResource(
              MAPPING_HOME + SolidifyConstants.URL_SEP + indexDefinition.getIndexSettings());
      String jsonSettings = FileTool.toString(settingsResource.getInputStream());
      indexSettings.setSettings(JSONTool.convertJson2Map(jsonSettings));
    }
  }
}
