/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Model - CacheListener.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.listener;

import javax.persistence.PostPersist;
import javax.persistence.PostRemove;
import javax.persistence.PostUpdate;

import org.springframework.context.ApplicationEventPublisher;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.message.CacheMessage;
import ch.unige.solidify.rest.JoinResource;
import ch.unige.solidify.rest.Resource;

public class CacheListener {

  @PostPersist
  /**
   * Publish a CacheMessage event that may be used to invalidate cache when a new relations between entities is created. The message is sent only
   * if the object is a JoinResource as there is no need to invalidate cache for new resources.
   */
  private void notifyPersist(Object object) {
    final ApplicationEventPublisher publisher = SolidifyEventPublisher.getPublisher();

    // publisher may be null when creating resources during initialization
    if (publisher != null && object instanceof JoinResource) {
      CacheMessage cacheMessage = ((JoinResource<?>) object).getCacheMessage();
      if (cacheMessage != null) {
        publisher.publishEvent(cacheMessage);
      }
    }
  }

  @PostUpdate
  @PostRemove
  /**
   * Publish a CacheMessage event that may be used to invalidate cache when object is updated or deleted. The message is sent if the object is a
   * simple Resource or a JoinResource as it may worth to invalidate cache in both cases.
   */
  private void notifyUpdate(Object object) {
    final ApplicationEventPublisher publisher = SolidifyEventPublisher.getPublisher();

    // publisher may be null when creating resources during initialization
    if (publisher != null) {
      CacheMessage cacheMessage = null;
      if (object instanceof Resource) {
        cacheMessage = ((Resource) object).getCacheMessage();
      } else if (object instanceof JoinResource) {
        cacheMessage = ((JoinResource<?>) object).getCacheMessage();
      }
      if (cacheMessage != null) {
        publisher.publishEvent(cacheMessage);
      }
    }
  }
}
